#ifndef c3d_H
#define c3d_H

#ifdef _WINDOWS
#define WINDOWS
#endif

#include "ac_plugin.h"
#include <math.h>
// START: Courtesy Nico De Ranter
#ifndef WINDOWS
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#define max(a,b) (((a) > (b)) ? (a) : (b))
#define min(a,b) (((a) < (b)) ? (a) : (b))
#define _finite(a) finite(a)
#define abs(a) fabs(a)
#endif
// END: Courtesy Nico De Ranter

// Planes
#define C3D_PLANE_X (0)
#define C3D_PLANE_Y (1)
#define C3D_PLANE_Z (2)

// Ray-tri intersects
#define C3D_ISECT_SUCCESS        0
#define C3D_ISECT_DEGENERATE_TRI 1
#define C3D_ISECT_NO_INTERSECT   2
#define C3D_ISECT_OUTSIDE_TRI    3

//---------------------------------------------------------------------------------------
// C3d : General 3d functions for fur support.
//
// Some functions (noted in comments above each function) were stringly derived
//    or outright copied from Andy Thaller's Extrude Along Path plugin for AC3D.
// This code was used in accordance with the GNU General Public License.
// For more information on Andy Thaller's work, see http://www.ph.tum.de/~thaller/extrude/.
//
class C3d
{

public:

   // Calculate the area of a parallelogram
   static float AreaOfParallelogram( Vertex *v1, Vertex *v2, Vertex *v3 )
   {
      // Calculate vector lengths
      Vertex l1, l2, cross;
      l1.x = v2->x - v1->x;
      l1.y = v2->y - v1->y;
      l1.z = v2->z - v1->z;
      l2.x = v3->x - v1->x;
      l2.y = v3->y - v1->y;
      l2.z = v3->z - v1->z;
      // Get the cross product of the vector lengths
      cross.x = pow(    l1.y * l2.z - l2.y * l1.z  , 2 );
      cross.y = pow( -( l1.x * l2.z - l2.x * l1.z ), 2 );
      cross.z = pow(    l1.x * l2.y - l2.x * l1.y  , 2 );
      // Calculate the magnitude
      cross.x = pow( cross.x, 2 );
      cross.y = pow( cross.y, 2 );
      cross.z = pow( cross.z, 2 );
      // Return the area
      return sqrtf( cross.x + cross.y + cross.z );
   }

   // Calculate the area of a triangle
   static float AreaOfTriangle( Vertex *v1, Vertex *v2, Vertex *v3 )
   {
      return ( AreaOfParallelogram( v1, v2, v3 ) * 0.5f );
   }

   // Return the "dominant" plane along which a normal lies
   static int GetDominantPlaneFromNormal( Point3 &norm )
   {
      if( fabs( norm.y ) > fabs( norm.x ))
      {
         if( fabs( norm.z ) > fabs( norm.y ))
         {
            return C3D_PLANE_Z;
         }
         else
         {
            return C3D_PLANE_Y;
         }
      }
      else
      {
         if( fabs( norm.z ) > fabs( norm.x ))
         {
            return C3D_PLANE_Z;
         }
         else
         {
            return C3D_PLANE_X;
         }
      }
   }

   // Cross product
   static Vertex CrossProduct( Vertex &v1, Vertex &v2 )
   {
      Vertex out;
      out.x = v1.y * v2.z - v1.z * v2.y;
      out.y = ( v1.z * v2.x - v1.x * v2.z );
      out.z = v1.x * v2.y - v1.y * v2.x;
      return out;
   }

   // Dot product
   static float DotProduct( Vertex &v1, Vertex &v2 )
   {
      return v1.x * v2.x + v1.y * v2.y + v1.z * v2.z;
   }

   // Intersect a triange with a ray, return the collision point in outI
   static int RayTriangleIntersect( Vertex *raystart, Vertex *raydir, Vertex triangle[3], Vertex *outI )
   {

      // Get triangle edge vectors and plane normal
      Vertex u;
      u.x = triangle[1].x - triangle[0].x;
      u.y = triangle[1].y - triangle[0].y;
      u.z = triangle[1].z - triangle[0].z;
      Vertex v;
      v.x = triangle[2].x - triangle[0].x;
      v.y = triangle[2].y - triangle[0].y;
      v.z = triangle[2].z - triangle[0].z;
      Vertex n = C3d::CrossProduct( u, v );

      Vertex w0;
      w0.x = raystart->x - triangle[0].x;
      w0.y = raystart->y - triangle[0].y;
      w0.z = raystart->z - triangle[0].z;
      float a = -( C3d::DotProduct( n, w0 ));
      float b = C3d::DotProduct( n, *raydir );
      if( fabs( b ) < 0.0000001f )
      {
         // ray is parallel to triangle plane
         if( a == 0 )
         {
            // ray lies in triangle plane
         }
         else
         {
            // ray disjoint from plane
         }
         return C3D_ISECT_DEGENERATE_TRI;
      }

      // get intersect point of ray with triangle plane
      float r = a / b;
      if( r < 0.0 )
      {
         // ray goes away from triangle
         return C3D_ISECT_NO_INTERSECT;
         // for a segment, also test if (r > 1.0) => no intersect
      }

      Vertex dirmag;
      dirmag.x = raydir->x * r;
      dirmag.y = raydir->y * r;
      dirmag.z = raydir->z * r;
      // intersect point of ray and plane
      outI->x = raystart->x + dirmag.x;
      outI->y = raystart->y + dirmag.y;
      outI->z = raystart->z + dirmag.z;

      // is intersect point in triangle
      float uu, uv, vv, wu, wv, D;
      uu = C3d::DotProduct( u, u );
      uv = C3d::DotProduct( u, v );
      vv = C3d::DotProduct( v, v );
      Vertex w;
      w.x = outI->x - triangle[0].x;
      w.y = outI->y - triangle[0].y;
      w.z = outI->z - triangle[0].z;
      wu = C3d::DotProduct( w, u );
      wv = C3d::DotProduct( w, v );
      D = uv * uv - uu * vv;

      // get and test parametric coords
      float s, t;
      s = ( uv * wv - vv * wu ) / D;
      if( s < 0.0 || s > 1.0 )
      {
         // I is outside T
         return C3D_ISECT_OUTSIDE_TRI;
      }
      t = ( uv * wu - uu * wv ) / D;
      if( t < 0.0 || ( s + t ) > 1.0 )
      {
         // I is outside T
         return C3D_ISECT_OUTSIDE_TRI;
      }

      return C3D_ISECT_SUCCESS;

   }

   // Calculate barycentric coordinates for a point
   static void BarycentricCoords( Vertex &point, Vertex &x1, Vertex &x2, Vertex &x3, Vertex &bcoords, float *baryover )
   {

      // Get the area of the triangle
      float a = C3d::AreaOfTriangle( &x1, &x2, &x3 );
      // Compute the barycentrics.
      // X = U, Y = V, W = Z
      bcoords.x = C3d::AreaOfTriangle( &x2, &x3, &point ) / a;
      bcoords.y = C3d::AreaOfTriangle( &x3, &x1, &point ) / a;
      bcoords.z = C3d::AreaOfTriangle( &x1, &x2, &point ) / a;

      if( 1.0f - ( bcoords.x + bcoords.y + bcoords.z ) - 0.000001f < 0.0f )
      {
         // Calculate the "normalized" barycentrics
         float barytot = bcoords.x + bcoords.y + bcoords.z;
         if( baryover != NULL )
         {
            *baryover = barytot - 1.0f;
         }
         bcoords.x /= barytot;
         bcoords.y /= barytot;
         bcoords.z /= barytot;
      }
      else
      {
         if( baryover != NULL )
         {
            *baryover = 0.0f;
         }
      }

   }

   // Interpolate normal from a point within a triangle
   static void InterpolateNormal( Vertex &point, SVertex *x1, SVertex *x2, SVertex *x3, Vertex &pointNormal )
   {
      Vertex bary;
      float baryover;
      C3d::BarycentricCoords( point, *( x1->v ), *( x2->v ), *( x3->v ), bary, &baryover );
      float baryl = bary.x + bary.y + bary.z;
      pointNormal.x = ( bary.x * x1->normal.x ) + ( bary.y * x2->normal.x ) + ( bary.z * x3->normal.x );
      pointNormal.y = ( bary.x * x1->normal.y ) + ( bary.y * x2->normal.y ) + ( bary.z * x3->normal.y );
      pointNormal.z = ( bary.x * x1->normal.z ) + ( bary.y * x2->normal.z ) + ( bary.z * x3->normal.z );
   }

   // Vector product
   // NOTE: This code is an adaptation of function(s) in Andy Thaller's Extrude Along Path plugin.
   static float VectorProduct( Vertex *vec1, Vertex *vec2 )
   {
      return vec1->x * vec2->x + vec1->y * vec2->y + vec1->z * vec2->z;
   }

   // Vector length
   // NOTE: This code is an adaptation of function(s) in Andy Thaller's Extrude Along Path plugin.
   static float VectorLength( Vertex *vec )
   {
      return sqrtf( vec->x * vec->x + vec->y * vec->y + vec->z * vec->z );
   }

   // Cosine angle between direction vectors
   // NOTE: This code is an adaptation of function(s) in Andy Thaller's Extrude Along Path plugin.
   static float VectorCosAngle( Vertex *vec1, Vertex *vec2 )
   {
      float l1 = C3d::VectorLength( vec1 );
      float l2 = C3d::VectorLength( vec2 );
      if( l1 == 0.0f || l2 == 0.0f )
      {
         return 0.0f;
      }
      return C3d::VectorProduct( vec1, vec2 ) / ( l1 * l2 );
   }

   // Vector axis
   // NOTE: This code is an adaptation of function(s) in Andy Thaller's Extrude Along Path plugin.
   static void VectorAxis( Vertex *vec1, Vertex *vec2, Vertex *axis )
   {
      float l;
      if( !axis )
      {
         return;
      }
      CROSSPRODUCT( vec1, vec2, axis );
      l = C3d::VectorLength( axis );
      if( l == 0.0f )
      {
         return;
      }
      axis->x /= l;
      axis->y /= l;
      axis->z /= l;
   }

   // Rotate a vertex
   // NOTE: This code is an adaptation of function(s) in Andy Thaller's Extrude Along Path plugin.
   static void RotateVertex( Vertex *v, Vertex *center, Vertex *axis, float C )
   {
      float S;
      float a = axis->x, b = axis->y, c = axis->z;
      float x0 = center->x, y0 = center->y, z0 = center->z;
      if( C < 0.0f )
      {
         C = 0.0;
      }
      if( C > 1.0f )
      {
         C = 1.0f;
      }
      S = sqrtf( 1.0f - C * C );
      float x= v->x - x0, y = v->y - y0, z = v->z - z0;
      v->x = (( 1 - C ) * a * a + C     ) * x + (( 1 -C ) * b * a - c * S ) * y + (( 1 - C ) * c * a + b * S ) * z + x0;
      v->y = (( 1 - C ) * a * b + c * S ) * x + (( 1 -C ) * b * b + C     ) * y + (( 1 - C ) * c * b - a * S ) * z + y0;
      v->z = (( 1 - C ) * a * c - b * S ) * x + (( 1 -C ) * b * c + a * S ) * y + (( 1 - C ) * c * c + C     ) * z + z0;
   }

   // NOTE: This code is an adaptation of function(s) in Andy Thaller's Extrude Along Path plugin.
   static int eap_matrix_xlines(double *d, int l1, int l2)
   {
      int i,w=4;
      double ftmp;

      for (i= 0; i<w; i++) {
         ftmp= d[l1*w+i];
         d[l1*w+i]= d[l2*w+i];
         d[l2*w+i]= ftmp;
      }
      return 1;
   }

   // NOTE: This code is an adaptation of function(s) in Andy Thaller's Extrude Along Path plugin.
   static int eap_calc_utri(double *d)
   {
      int i,j,w=4,h=3,l,k,itmp, l1, l2;
      double min, ftmp;
      int done[3] = { -1, -1, -1 };

      for (i= 0; i<w-1; i++) {
         /* find the line l with the min value in column i:) */
         min= -1; l= -1;
         for (j= 0; j<h; j++) {
            ftmp= fabs(d[i+j*w]);
            if ((min<0 || ftmp<min) && ftmp>0 && done[j]<0) { min= ftmp; l= j; }
         }
         if (l>=0) {
            done[l]= i; /* Normalize line l: */
            ftmp= d[l*w+i]; for (k= 0; k<w; k++) d[l*w+k]/= -ftmp;
            for (j= 0; j<h; j++) /* for every line except the l-th: */
               if (j!=l) {
                  /* remeber the factor to multiply the l-th line with: */
                  ftmp= d[j*w+i];
                  for (k= 0; k<w; k++)
                     d[j*w+k]= d[j*w+k]*d[l*w+i]- ftmp* d[l*w+k];
               }
         }
      }
      for (j= 0; j<h; j++) { /* sort the matrix lines: */
         l1= j; l2= -1;
         while (l2!=j) if (done[j]>=0) {
            l2= done[j];
            eap_matrix_xlines(d,l1,l2);
            itmp= done[l1]; done[l1]= done[l2]; done[l2]= itmp;
         } else l2= l1;
      }
      return 1;
   }

   // Intersect two lines
   //
   // Written by A. Thaller, with slight modifications
   //
   // line1 contains a1 and has direction b1= c1-a1
   // line2 contains a2 and has direction b2= c2-a2
   //
   // returns the point X of intersection, or the point in the center of the
   // shortest connection between line1 and line2
   static void vtx_line_intersection( Vertex &vm, Vertex &a1, Vertex &c1, Vertex &a2, Vertex &c2 )
   {
      double cosa, d;
      double m[12], r,s,t;
      Vertex b1= c1, b2= c2, v= a2, b1b2, a1a2= a2, p1,p2;
      /* b1 and b2 are points of the respective lines, make them directions: */
      /* b1-= a1; b2-= a2 */
      SUBTRACTPOINTS(&a1,&b1); SUBTRACTPOINTS(&a2,&b2);
      SUBTRACTPOINTS(&a1,&v); SUBTRACTPOINTS(&a1,&a1a2);

      /* we cannot operate on parallel lines: */
      cosa= C3d::VectorCosAngle( &b1, &b2 );
      if (cosa==1.0)
      {
         memcpy( &vm, &c1, sizeof( Vertex ));
         return;
      }
      CROSSPRODUCT(&b1,&b2,&b1b2);
      d= fabs(DOTPRODUCT(&v,&b1b2));

      /* setup the matrix for the equation system and unify (solve) it */
      m[ 0]= b1.x; m[ 1]= -b2.x; m[ 2]= b1b2.x; m[ 3]= a1a2.x;
      m[ 4]= b1.y; m[ 5]= -b2.y; m[ 6]= b1b2.y; m[ 7]= a1a2.y;
      m[ 8]= b1.z; m[ 9]= -b2.z; m[10]= b1b2.z; m[11]= a1a2.z;
      eap_calc_utri(m); r= -m[3]; s= -m[7]; t= -m[11];

      /* calculate the two points which are nearest to each other: */
      p1= b1; MULTIPLYPOINT(&p1,r); ADDPOINTS(&a1,&p1);
      p2= b2; MULTIPLYPOINT(&p2,s); ADDPOINTS(&a2,&p2);

      /* return the midpoint of those two: */
      MIDPOINT(&p1,&p1,&vm);
   }

};
//---------------------------------------------------------------------------------------

#endif
