#include "cgreebler.h"
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <math.h>

#define DRAND ((double)rand() / ((double)(RAND_MAX)+(double)(1)))

//----------------------------------------------------------------------
void ddhGetCustomObjects( List **out, ACObject *obj )
{
   for( List *po = ac_object_get_childrenlist( obj ); po != NULL; po = po->next )
   {
      ACObject *o = ( ACObject * )( po->data );
      if( ac_object_get_childrenlist( o ) != NULL )
      {
         ddhGetCustomObjects( out, o );
      }
      else
      {
         char *p = ac_object_get_data( o );
         bool newline = true;
         if( p == NULL )
         {
            continue;
         }
         while( p[0] != '\0' )
         {
            if( newline )
            {
               if( strncmp( p, "//CUSTOM_GREEBLE", 16 ) == 0 )
               {
                  list_add_item_head( out, o );
                  break;
               }
            }
            newline = ( p[0] == 10 || p[0] == 13 );
            ++p;
         }
      }
   }
}
//----------------------------------------------------------------------
// 2D line intersect
bool isect_2d( Point2 pa, Point2 pb, Point2 pc, Point2 pd, Point2 *out )
{
   float f = (( pa.y - pc.y ) * ( pd.x - pc.x )) - (( pa.x - pc.x ) * ( pd.y - pc.y ));
   float g = (( pb.x - pa.x ) * ( pd.y - pc.y )) - (( pb.y - pa.y ) * ( pd.x - pc.x ));
   float h = (( pa.y - pc.y ) * ( pb.x - pa.x )) - (( pa.x - pc.x ) * ( pb.y - pc.y ));
   if( g == 0 )
   {
      out->x = pa.x;
      out->y = pa.y;
   }
   float r = f / g;
   float s = h / g;
   out->x = pa.x + r * ( pb.x - pa.x );
   out->y = pa.y + r * ( pb.y - pa.y );
   return !( r < 0 || r > 1 || s < 0 || s > 1 );
}
//----------------------------------------------------------------------
CGreebler::CGreebler()
{

   mGreebles = NULL;
   mNumGreebles = 0;
   mRandSeed = ( unsigned )time( NULL );
   mSmallPolySize = 0.0005f;
   mObjKeepSeparate = true;

   mDoBevel = true;
   mBevRandom = true;
   mBevRandomPct = 0.8f;
   mBevDistancePctMin = 0.7f;
   mBevDistancePctMax = 1.0f;
   mBevIndentPctMin = 0.01f;
   mBevIndentPctMax = 0.1f;
   mBevGridPct = 1.0f;

   mDoGreeble = true;
   mGreRandom = true;
   mGreRandomPct = 0.8f;
   mGreNumPerSurfaceMin = 1;
   mGreNumPerSurfaceMax = 4;
   mGreHeightPctMin = 0.2f;
   mGreHeightPctMax = 1.0f;
   mGreSizePctMin = 0.1f;
   mGreSizePctMax = 0.45f;
   mGreRotIncrDegrees = 90.0f;
   mGreCustNoConform = true;

}
//----------------------------------------------------------------------
CGreebler::~CGreebler()
{
   ClearGreebleCollection();
}
//----------------------------------------------------------------------
void CGreebler::ClearGreebleCollection()
{
   if( mNumGreebles != 0 )
   {
      delete[] mGreebles;
      mGreebles = NULL;
      mNumGreebles = 0;
   }
}
//----------------------------------------------------------------------
void acvlist2arr( List *vlist, int &numverts, float *x, float *y, float *z )
{
   for( List *pv = vlist; pv != NULL; pv = pv->next )
   {
      SVertex *v = ( SVertex * )( pv->data );
      x[numverts] = v->v->x;
      y[numverts] = v->v->y;
      z[numverts] = v->v->z;
      ++numverts;
   }
}
//----------------------------------------------------------------------
List *CGreebler::Generate( bool add_undoable )
{

   if( !mDoBevel && !mDoGreeble )
   {
      message_dialog( "greeble : neither beveling or greebling was selected; nothing to do" );
   }

   // Grab a list of selected surfaces
   List *selsurf = ac_selection_get_whole_surfaces_all();
   if( selsurf == NULL )
   {
      message_dialog( "greeble : no whole surfaces were selected." );
      return NULL;
   }

   // Seed randomizer
   srand( this->mRandSeed );

   // List of beveled surfaces
   List *bevsurfs = NULL;

   // New vertices
   List *newverts = NULL;
   // New surfaces
   List *newsurfs = NULL;

   // New objects
   // Only used for mObjKeepSeparate option
   List *newobjs = NULL;

   // NOTE: These definitions are based on maximum number of sides we can greeble (4).
   float x[4], y[4], z[4];

   // Begin beveling
   int nontriquad_skipped = 0;
   int small_skipped = 0;
   float seglen[4];
   set_progress_bar( 0 );
   int numsurfs = list_count( selsurf );
   if( mDoBevel )
   {
      display_message( "Beveling %d surfaces", numsurfs );
   }
   else
   {
      display_message( "Evaluating %d surfaces", numsurfs );
   }
   int ctr = 0;
   double total_work;
   if( mDoGreeble )
   {
      total_work = 50;
   }
   else
   {
      total_work = 100;
   }
   for( List *ps = selsurf; ps != NULL; ps = ps->next )
   {

      Surface *currsurf = ( Surface * )( ps->data );
      set_progress_bar(( int )((( double )ctr / ( double )numsurfs ) * total_work ));
      ++ctr;

      // Skip non-quad polys
      if( currsurf->numvert != 4 && currsurf->numvert != 3 )
      {
         ++nontriquad_skipped;
         continue;
      }

      // Detect whether this is a "small poly"
      bool issmall = false;
      List *psv = currsurf->vertlist;
      Vertex *surfverts[7];
      for( int ix = 0; ix < currsurf->numvert; ++ix )
      {
         SVertex *sv = ( SVertex * )( psv->data );
         surfverts[ix] = sv->v;
         psv = psv->next;
      }
      // Triangle
      if( currsurf->numvert == 3 )
      {
         // Set the final (4th-5th) verts so we don't have to to do a conditional
         surfverts[3] = surfverts[0];
         surfverts[4] = surfverts[1];
         for( int ix = 0; ix < 3; ++ix )
         {
            if( line_seg_point_dist_3d( surfverts[ix+1]->x, surfverts[ix+1]->y, surfverts[ix+1]->z,
                                        surfverts[ix+2]->x, surfverts[ix+2]->y, surfverts[ix+2]->z,
                                        surfverts[ix]->x, surfverts[ix]->y, surfverts[ix]->z ) < mSmallPolySize )
            {
               issmall = true;
               break;
            }
         }
      }
      else if( currsurf->numvert == 4 )
      {
         // Set the final (5th-6th-7th) verts so we don't have to to do a conditional
         surfverts[4] = surfverts[0];
         surfverts[5] = surfverts[1];
         surfverts[6] = surfverts[2];
         for( int ix = 0; ix < 4; ++ix )
         {
            if( line_seg_point_dist_3d( surfverts[ix+1]->x, surfverts[ix+1]->y, surfverts[ix+1]->z,
                                        surfverts[ix+2]->x, surfverts[ix+2]->y, surfverts[ix+2]->z,
                                        surfverts[ix]->x, surfverts[ix]->y, surfverts[ix]->z ) < mSmallPolySize ||
                line_seg_point_dist_3d( surfverts[ix+2]->x, surfverts[ix+2]->y, surfverts[ix+2]->z,
                                        surfverts[ix+3]->x, surfverts[ix+3]->y, surfverts[ix+3]->z,
                                        surfverts[ix]->x, surfverts[ix]->y, surfverts[ix]->z ) < mSmallPolySize )
            {
               issmall = true;
               break;
            }
         }
      }
      if( issmall )
      {
         // Too small - skip it
         ++small_skipped;
         continue;
      }

      // Grab the vertices of the surface into arrays for x/y/z
      int numverts = 0;
      acvlist2arr( currsurf->vertlist, numverts, x, y, z );

      // Bevel
      bool didbevel = false;
      if( mDoBevel )
      {
         // Randomly skip bevels
         if( mBevRandom && DRAND > mBevRandomPct )
         {
            // Skip
         }
         else
         {
            // Split surface of quads
            bool dogrid = ( DRAND < mBevGridPct );
            if( currsurf->numvert == 4 && dogrid )
            {
               double drand = DRAND;
               if( drand < 0.333333333333 )
               {
                  // Randomly shift the verts so the splits aren't always in the same spots
                  float txarr[4], tyarr[4], tzarr[4];
                  double drand2 = DRAND;
                  if( drand2 < 0.5 )
                  {
                     txarr[0] = x[0];
                     tyarr[0] = y[0];
                     tzarr[0] = z[0];
                     x[0] = x[1];
                     y[0] = y[1];
                     z[0] = z[1];
                     x[1] = x[2];
                     y[1] = y[2];
                     z[1] = z[2];
                     x[2] = x[3];
                     y[2] = y[3];
                     z[2] = z[3];
                     x[3] = txarr[0];
                     y[3] = tyarr[0];
                     z[3] = tzarr[0];
                  }
                  // Two bevels
                  txarr[0] = x[0];
                  tyarr[0] = y[0];
                  tzarr[0] = z[0];
                  txarr[1] = x[1];
                  tyarr[1] = y[1];
                  tzarr[1] = z[1];
                  txarr[2] = x[1] - ( x[1] - x[2] ) * 0.5f;
                  tyarr[2] = y[1] - ( y[1] - y[2] ) * 0.5f;
                  tzarr[2] = z[1] - ( z[1] - z[2] ) * 0.5f;
                  txarr[3] = x[0] - ( x[0] - x[3] ) * 0.5f;
                  tyarr[3] = y[0] - ( y[0] - y[3] ) * 0.5f;
                  tzarr[3] = z[0] - ( z[0] - z[3] ) * 0.5f;
                  list_add_item_head( &bevsurfs, BevelSurface( &newobjs, &newsurfs, &newverts, currsurf, numverts, txarr, tyarr, tzarr ));
                  txarr[0] = x[0] - ( x[0] - x[3] ) * 0.5f;
                  tyarr[0] = y[0] - ( y[0] - y[3] ) * 0.5f;
                  tzarr[0] = z[0] - ( z[0] - z[3] ) * 0.5f;
                  txarr[1] = x[1] - ( x[1] - x[2] ) * 0.5f;
                  tyarr[1] = y[1] - ( y[1] - y[2] ) * 0.5f;
                  tzarr[1] = z[1] - ( z[1] - z[2] ) * 0.5f;
                  txarr[2] = x[2];
                  tyarr[2] = y[2];
                  tzarr[2] = z[2];
                  txarr[3] = x[3];
                  tyarr[3] = y[3];
                  tzarr[3] = z[3];
                  list_add_item_head( &bevsurfs, BevelSurface( &newobjs, &newsurfs, &newverts, currsurf, numverts, txarr, tyarr, tzarr ));
               }
               else if( drand < 0.666666666666 )
               {
                  // Randomly shift the verts so the splits aren't always in the same spots
                  float txarr[4], tyarr[4], tzarr[4];
                  double drand2 = DRAND;
                  if( drand2 < 0.25 )
                  {
                     txarr[0] = x[0];
                     tyarr[0] = y[0];
                     tzarr[0] = z[0];
                     x[0] = x[1];
                     y[0] = y[1];
                     z[0] = z[1];
                     x[1] = x[2];
                     y[1] = y[2];
                     z[1] = z[2];
                     x[2] = x[3];
                     y[2] = y[3];
                     z[2] = z[3];
                     x[3] = txarr[0];
                     y[3] = tyarr[0];
                     z[3] = tzarr[0];
                  }
                  else if( drand2 < 0.5 )
                  {
                     txarr[0] = x[0];
                     tyarr[0] = y[0];
                     tzarr[0] = z[0];
                     txarr[1] = x[1];
                     tyarr[1] = y[1];
                     tzarr[1] = z[1];
                     x[0] = x[2];
                     y[0] = y[2];
                     z[0] = z[2];
                     x[1] = x[3];
                     y[1] = y[3];
                     z[1] = z[3];
                     x[2] = txarr[0];
                     y[2] = tyarr[0];
                     z[2] = tzarr[0];
                     x[3] = txarr[1];
                     y[3] = tyarr[1];
                     z[3] = tzarr[1];
                  }
                  else if( drand2 < 0.75 )
                  {
                     txarr[0] = x[2];
                     tyarr[0] = y[2];
                     tzarr[0] = z[2];
                     x[2] = x[1];
                     y[2] = y[1];
                     z[2] = z[1];
                     x[1] = x[0];
                     y[1] = y[0];
                     z[1] = z[0];
                     x[0] = x[3];
                     y[0] = y[3];
                     z[0] = z[3];
                     x[3] = txarr[0];
                     y[3] = tyarr[0];
                     z[3] = tzarr[0];
                  }
                  // Three bevels
                  Vertex polySide[3];
                  Vertex polyCenter;
                  polySide[0].x = x[1] - ( x[1] - x[2] ) * 0.5f;
                  polySide[0].y = y[1] - ( y[1] - y[2] ) * 0.5f;
                  polySide[0].z = z[1] - ( z[1] - z[2] ) * 0.5f;
                  polySide[1].x = x[2] - ( x[2] - x[3] ) * 0.5f;
                  polySide[1].y = y[2] - ( y[2] - y[3] ) * 0.5f;
                  polySide[1].z = z[2] - ( z[2] - z[3] ) * 0.5f;
                  polySide[2].x = x[0] - ( x[0] - x[3] ) * 0.5f;
                  polySide[2].y = y[0] - ( y[0] - y[3] ) * 0.5f;
                  polySide[2].z = z[0] - ( z[0] - z[3] ) * 0.5f;
                  polyCenter.x = x[0] - ( x[0] - x[2] ) * 0.5f;
                  polyCenter.y = y[0] - ( y[0] - y[2] ) * 0.5f;
                  polyCenter.z = z[0] - ( z[0] - z[2] ) * 0.5f;
                  polyCenter.x = (( x[1] - ( x[1] - x[3] ) * 0.5f ) + polyCenter.x ) * 0.5f;
                  polyCenter.y = (( y[1] - ( y[1] - y[3] ) * 0.5f ) + polyCenter.y ) * 0.5f;
                  polyCenter.z = (( z[1] - ( z[1] - z[3] ) * 0.5f ) + polyCenter.z ) * 0.5f;
                  txarr[0] = x[0];
                  tyarr[0] = y[0];
                  tzarr[0] = z[0];
                  txarr[1] = x[1];
                  tyarr[1] = y[1];
                  tzarr[1] = z[1];
                  txarr[2] = polySide[0].x;
                  tyarr[2] = polySide[0].y;
                  tzarr[2] = polySide[0].z;
                  txarr[3] = polySide[2].x;
                  tyarr[3] = polySide[2].y;
                  tzarr[3] = polySide[2].z;
                  list_add_item_head( &bevsurfs, BevelSurface( &newobjs, &newsurfs, &newverts, currsurf, numverts, txarr, tyarr, tzarr ));
                  txarr[0] = x[2];
                  tyarr[0] = y[2];
                  tzarr[0] = z[2];
                  txarr[1] = polySide[1].x;
                  tyarr[1] = polySide[1].y;
                  tzarr[1] = polySide[1].z;
                  txarr[2] = polyCenter.x;
                  tyarr[2] = polyCenter.y;
                  tzarr[2] = polyCenter.z;
                  txarr[3] = polySide[0].x;
                  tyarr[3] = polySide[0].y;
                  tzarr[3] = polySide[0].z;
                  list_add_item_head( &bevsurfs, BevelSurface( &newobjs, &newsurfs, &newverts, currsurf, numverts, txarr, tyarr, tzarr ));
                  txarr[0] = x[3];
                  tyarr[0] = y[3];
                  tzarr[0] = z[3];
                  txarr[1] = polySide[2].x;
                  tyarr[1] = polySide[2].y;
                  tzarr[1] = polySide[2].z;
                  txarr[2] = polyCenter.x;
                  tyarr[2] = polyCenter.y;
                  tzarr[2] = polyCenter.z;
                  txarr[3] = polySide[1].x;
                  tyarr[3] = polySide[1].y;
                  tzarr[3] = polySide[1].z;
                  list_add_item_head( &bevsurfs, BevelSurface( &newobjs, &newsurfs, &newverts, currsurf, numverts, txarr, tyarr, tzarr ));
               }
               else
               {
                  // Four bevels
                  Vertex polySide[4];
                  Vertex polyCenter;
                  polySide[0].x = x[0] - ( x[0] - x[1] ) * 0.5f;
                  polySide[0].y = y[0] - ( y[0] - y[1] ) * 0.5f;
                  polySide[0].z = z[0] - ( z[0] - z[1] ) * 0.5f;
                  polySide[1].x = x[1] - ( x[1] - x[2] ) * 0.5f;
                  polySide[1].y = y[1] - ( y[1] - y[2] ) * 0.5f;
                  polySide[1].z = z[1] - ( z[1] - z[2] ) * 0.5f;
                  polySide[2].x = x[2] - ( x[2] - x[3] ) * 0.5f;
                  polySide[2].y = y[2] - ( y[2] - y[3] ) * 0.5f;
                  polySide[2].z = z[2] - ( z[2] - z[3] ) * 0.5f;
                  polySide[3].x = x[0] - ( x[0] - x[3] ) * 0.5f;
                  polySide[3].y = y[0] - ( y[0] - y[3] ) * 0.5f;
                  polySide[3].z = z[0] - ( z[0] - z[3] ) * 0.5f;
                  polyCenter.x = x[0] - ( x[0] - x[2] ) * 0.5f;
                  polyCenter.y = y[0] - ( y[0] - y[2] ) * 0.5f;
                  polyCenter.z = z[0] - ( z[0] - z[2] ) * 0.5f;
                  polyCenter.x = (( x[1] - ( x[1] - x[3] ) * 0.5f ) + polyCenter.x ) * 0.5f;
                  polyCenter.y = (( y[1] - ( y[1] - y[3] ) * 0.5f ) + polyCenter.y ) * 0.5f;
                  polyCenter.z = (( z[1] - ( z[1] - z[3] ) * 0.5f ) + polyCenter.z ) * 0.5f;
                  float txarr[4], tyarr[4], tzarr[4];
                  txarr[0] = x[0];
                  tyarr[0] = y[0];
                  tzarr[0] = z[0];
                  txarr[1] = polySide[0].x;
                  tyarr[1] = polySide[0].y;
                  tzarr[1] = polySide[0].z;
                  txarr[2] = polyCenter.x;
                  tyarr[2] = polyCenter.y;
                  tzarr[2] = polyCenter.z;
                  txarr[3] = polySide[3].x;
                  tyarr[3] = polySide[3].y;
                  tzarr[3] = polySide[3].z;
                  list_add_item_head( &bevsurfs, BevelSurface( &newobjs, &newsurfs, &newverts, currsurf, numverts, txarr, tyarr, tzarr ));
                  txarr[0] = x[1];
                  tyarr[0] = y[1];
                  tzarr[0] = z[1];
                  txarr[1] = polySide[1].x;
                  tyarr[1] = polySide[1].y;
                  tzarr[1] = polySide[1].z;
                  txarr[2] = polyCenter.x;
                  tyarr[2] = polyCenter.y;
                  tzarr[2] = polyCenter.z;
                  txarr[3] = polySide[0].x;
                  tyarr[3] = polySide[0].y;
                  tzarr[3] = polySide[0].z;
                  list_add_item_head( &bevsurfs, BevelSurface( &newobjs, &newsurfs, &newverts, currsurf, numverts, txarr, tyarr, tzarr ));
                  txarr[0] = x[2];
                  tyarr[0] = y[2];
                  tzarr[0] = z[2];
                  txarr[1] = polySide[2].x;
                  tyarr[1] = polySide[2].y;
                  tzarr[1] = polySide[2].z;
                  txarr[2] = polyCenter.x;
                  tyarr[2] = polyCenter.y;
                  tzarr[2] = polyCenter.z;
                  txarr[3] = polySide[1].x;
                  tyarr[3] = polySide[1].y;
                  tzarr[3] = polySide[1].z;
                  list_add_item_head( &bevsurfs, BevelSurface( &newobjs, &newsurfs, &newverts, currsurf, numverts, txarr, tyarr, tzarr ));
                  txarr[0] = x[3];
                  tyarr[0] = y[3];
                  tzarr[0] = z[3];
                  txarr[1] = polySide[3].x;
                  tyarr[1] = polySide[3].y;
                  tzarr[1] = polySide[3].z;
                  txarr[2] = polyCenter.x;
                  tyarr[2] = polyCenter.y;
                  tzarr[2] = polyCenter.z;
                  txarr[3] = polySide[2].x;
                  tyarr[3] = polySide[2].y;
                  tzarr[3] = polySide[2].z;
                  list_add_item_head( &bevsurfs, BevelSurface( &newobjs, &newsurfs, &newverts, currsurf, numverts, txarr, tyarr, tzarr ));
               }
            }
            else
            {
               list_add_item_head( &bevsurfs, BevelSurface( &newobjs, &newsurfs, &newverts, currsurf, numverts, x, y, z ));
            }
            didbevel = true;
         }
      }
      if( !didbevel )
      {
         list_add_item_head( &bevsurfs, currsurf );
      }

   }

   // Begin greebling
   if( mDoGreeble )
   {

      numsurfs = list_count( bevsurfs );
      display_message( "Greebling %d surfaces", numsurfs );
      ctr = 0;

      // Gather a list of custom greebles
      int num_greebles = 0;
      List *po = NULL;
      if( mUseCustom )
      {
         ddhGetCustomObjects( &po, ac_get_world() );
         num_greebles = list_count( po );
      }

      // Add standards
      if( num_greebles == 0 || mUseStandard )
      {
         num_greebles += 4;
      }

      CGreeble *greebles = new CGreeble[num_greebles];

      // Load the standard greeble collection
      int ctr = 0;
      if( num_greebles == 0 || mUseStandard )
      {
         greebles[0].CreateCubic();
         greebles[1].CreateL();
         greebles[2].CreateT();
         greebles[3].CreateIII();
         ctr = 4;
      }

      // Load custom greebles
      for( List *pobj = po; pobj != NULL; pobj = pobj->next )
      {
         greebles[ctr].LoadFromObject(( ACObject * )( pobj->data ));
         if( greebles[ctr].mNumSurfs == 0 || greebles[ctr].mNumVerts == 0 )
         {
            // Ignore this greeble - no data
            --num_greebles;
         }
         else
         {
            ++ctr;
         }
      }

      // Cleanup list
      list_free( &po );

      for( List *ps = bevsurfs; ps != NULL; ps = ps->next )
      {

         Surface *currsurf = ( Surface * )( ps->data );
         set_progress_bar(( int )((( double )ctr / ( double )numsurfs ) * total_work ) + 50 );
         ++ctr;

         // Randomly skip greebling
         if( mGreRandom && DRAND > mGreRandomPct )
         {
            continue;
         }

         int numgreebles = min(( int )( DRAND * ( double )( mGreNumPerSurfaceMax + 1 - mGreNumPerSurfaceMin )) + mGreNumPerSurfaceMin, mGreNumPerSurfaceMax );

         // Prepare a target surface
         CTargetSurface tsurf;
         tsurf.PrepareSurface( currsurf );

         // Greeble
         for( int ix = 0; ix < numgreebles; ++ix )
         {
            CGreeble *g = &( greebles[( int )( DRAND * (( double )num_greebles ))] );
            GreebleSurface( &newobjs, &newsurfs, &newverts, &tsurf, g );
         }

      }

      // Get rid of greebles
      delete[] greebles;

   }

   ACObject *outobj = NULL;
   List *outobjlist = NULL;
   if( newverts != NULL )
   {
      if( mObjKeepSeparate )
      {
         // Everything in separate objects, grouped
         outobj = ac_new_group();
         ac_object_set_name( outobj, "greebles" );
         for( List *po = newobjs; po != NULL; po = po->next )
         {
            ACObject *thisobj = ( ACObject * )( po->data );
            list_add_item_head( &outobjlist, thisobj );
            //object_add_child( outobj, thisobj );
            object_reparent( thisobj, outobj );
            object_calc_normals_force( thisobj );
         }
         list_free( &newobjs );
         // Add the group
         if( add_undoable )
         {
            add_undoable_world_add_object( "greeble", outobj );
         }
         else
         {
            object_add_child( ac_get_world(), outobj );
         }
         // Add the group container to the front of the object list
         list_add_item_head( &outobjlist, outobj );
      }
      else
      {
         // Everything in one object
         outobj = ac_new_object();
         ac_object_set_name( outobj, "greeble" );
         object_add_vertices( outobj, newverts );
         object_add_surfaces( outobj, newsurfs );
         object_calc_normals_force( outobj );
         // Add the object to the object list
         list_add_item_head( &outobjlist, outobj );
         // Add the object list
         if( add_undoable )
         {
            add_undoable_world_add_objectlist( "greeble", outobjlist );
         }
         else
         {
            object_add_children( ac_get_world(), outobjlist );
         }
      }
   }

   // Free the beveled surfaces list
   list_free( &bevsurfs );
   list_free( &selsurf );

   if( nontriquad_skipped > 0 )
   {
      message_dialog( "greeble : %d polygons with less than 3 sides or more than 4 sides were skipped.", nontriquad_skipped );
   }
   if( small_skipped > 0 )
   {
      message_dialog( "greeble : %d polygons considered too small/thin (less than %g wide) were skipped.", small_skipped );
   }

   set_progress_bar( 0 );

   return outobjlist;

}
//----------------------------------------------------------------------
Surface *CGreebler::BevelSurface( List **newobjs, List **newsurfs, List **newverts, Surface *surf, int numverts, float xarr[], float yarr[], float zarr[] )
{

   // Bevel distance
   float bevdist = ( float )( DRAND * ( mBevDistancePctMax - mBevDistancePctMin ) + mBevDistancePctMin );

   // Calculate area of the polygon
   float area = polygon_area_2_3d( numverts, xarr, yarr, zarr );

   // Set the bevel height to the area divided by the number of sides.
   // This is fairly arbitrary, but gives a nice look that's proportional by surface area.
   float bevheight = max( 0.0001f, ( area / surf->numvert ) * bevdist * 4 );

   // Grab indent amount
   float bevindent = ( float )( DRAND * ( mBevIndentPctMax - mBevIndentPctMin ) + mBevIndentPctMin );

   // Constrain bevel distance
   bevindent = min( 100.0f, max( 0.0f, bevindent ));

   // Get the center of the polygon
   Vertex polyCenter;
   polygon_centroid_3d( numverts, xarr, yarr, zarr, &( polyCenter.x ), &( polyCenter.y ), &( polyCenter.z ));

   // Create the object
   ACObject *newobj = NULL;
   if( mObjKeepSeparate )
   {
      newobj = ac_new_object();
      ac_object_set_name( newobj, "bevel" );
      list_add_item_head( newobjs, newobj );
   }

   // Create the vertices to make the bevel
   List *varr = NULL;
   int ctr = 0;
   for( int ix = 0; ix < numverts; ++ix )
   {
      Vertex *newv = new_vertex_set( xarr[ix], yarr[ix], zarr[ix] );
      // Move the vertices "inward" toward the centroid
      newv->x = newv->x - ( newv->x - polyCenter.x ) * bevindent;
      newv->y = newv->y - ( newv->y - polyCenter.y ) * bevindent;
      newv->z = newv->z - ( newv->z - polyCenter.z ) * bevindent;
      list_add_item( &varr, newv );
      if( mObjKeepSeparate )
      {
         object_add_vertex_head( newobj, newv );
      }
      // Add vertex to final list
      list_add_item_head( newverts, newv );
      ++ctr;
   }

   // Create top of bevel
   List *varr2 = NULL;
   for( List *pv = varr; pv != NULL; pv = pv->next )
   {
      Vertex *v = ( Vertex * )( pv->data );
      Vertex *newv = new_vertex_set( v->x, v->y, v->z );
      newv->x = newv->x + surf->normal.x * bevheight;
      newv->y = newv->y + surf->normal.y * bevheight;
      newv->z = newv->z + surf->normal.z * bevheight;
      list_add_item( &varr2, newv );
      if( mObjKeepSeparate )
      {
         object_add_vertex_head( newobj, newv );
      }
      // Add vertex to final list
      list_add_item_head( newverts, newv );
      --ctr;
      if( ctr == 0 )
      {
         // All done - don't loop again or we'll add using our new verts
         break;
      }
   }

   Surface *s;
   Vertex *v1, *v2, *v3, *v4;

   // Add side surfaces
   List *vcur = varr;
   List *vcur2 = varr2;
   for( int ix = 0; ix < surf->numvert; ++ix )
   {
      s = new_surface();
      v1 = ( Vertex * )( vcur->data );
      v4 = ( Vertex * )( vcur2->data );
      if( ix == surf->numvert - 1 )
      {
         v2 = ( Vertex * )( varr->data );
         v3 = ( Vertex * )( varr2->data );
      }
      else
      {
         v2 = ( Vertex * )( vcur->next->data );
         v3 = ( Vertex * )( vcur2->next->data );
      }
      surface_add_vertex( s, v1, 0.0f, 1.0f );
      surface_add_vertex( s, v2, 1.0f, 1.0f );
      surface_add_vertex( s, v3, 1.0f, 0.0f );
      surface_add_vertex( s, v4, 0.0f, 0.0f );
      surface_copy_properties( surf, s );
      list_add_item_head( newsurfs, s );
      if( mObjKeepSeparate )
      {
         object_add_surface_head( newobj, s );
      }
      vcur = vcur->next;
      vcur2 = vcur2->next;
   }

   // Add top
   s = new_surface();
   List *psv_orig = surf->vertlist;
   for( List *pv = varr2; pv != NULL; pv = pv->next )
   {
      Vertex *v = ( Vertex * )( pv->data );
      // Copy the original surfaces' texture coordinates
      SVertex *sv = ( SVertex * )( psv_orig->data );
      surface_add_vertex( s, v, sv->tx, sv->ty );
      psv_orig = psv_orig->next;
   }
   surface_copy_properties( surf, s );
   list_add_item_head( newsurfs, s );
   if( mObjKeepSeparate )
   {
      object_add_surface_head( newobj, s );
   }

   // Copy the texture index
   if( mObjKeepSeparate )
   {
      ACObject *parentobj = object_of_surface( surf );
      if( ac_object_has_texture( parentobj ))
      {
         object_texture_set( newobj, ac_object_get_texture_index( parentobj ));
      }
   }

   // Return top surface
   return s;

}
//----------------------------------------------------------------------
bool point_in_poly( Point2 *poly_verts, int num_verts, Point2 *test_point )
{
   int cn = 0;
   for( int ix = 0; ix < num_verts; ++ix )
   {
      if((( poly_verts[ix].y <= test_point->y ) && ( poly_verts[ix+1].y > test_point->y )) ||
         (( poly_verts[ix].y > test_point->y ) && ( poly_verts[ix+1].y <= test_point->y )))
      {
         float vt = ( float )( test_point->y - poly_verts[ix].y ) / ( poly_verts[ix+1].y - poly_verts[ix].y );
         if( test_point->x < poly_verts[ix].x + vt * ( poly_verts[ix+1].x - poly_verts[ix].x ))
         {
            ++cn;
         }
      }
   }
   return (( cn & 1 ) == 1 );
}
//----------------------------------------------------------------------
void CGreebler::GreebleSurface( List **newobjs, List **newsurfs, List **newverts, CTargetSurface *surf, CGreeble *in_greeble )
{

   // NOTE: Much of the math in this section is based on the greeble being centered at x0z0,
   //       and with the base of the greeble at y0

   // Copy greeble geometry data into local object
   CGreObject o;
   o.CopyGreObjectFrom( in_greeble );

   // Rotate greeble about the Y axis
   float rotangle = DEGTORAD((( int )(( DRAND * 360.0 ) / mGreRotIncrDegrees ) * mGreRotIncrDegrees ));
   for( unsigned int ix = 0; ix < o.mNumVerts; ++ix )
   {
      CGreVertex *v = &( o.mVerts[ix] );
      float newx = v->v.x * cos( rotangle ) - v->v.z * sin( rotangle );
      float newz = v->v.x * sin( rotangle ) + v->v.z * cos( rotangle );
      v->v.x = newx;
      v->v.z = newz;
   }
   // Recalculate the center/min/max
   o.FindBounds();

   // bl,br,tl,tr Bottom left, bottom right, top left, and top right vertices in the target
   //             surface.  Just a shortcut.
   Vertex *br = &( surf->xform_points[0] );
   Vertex *tr = &( surf->xform_points[1] );
   Vertex *tl = &( surf->xform_points[2] );
   Vertex *bl = &( surf->xform_points[3] );

   // Whether to reverse the winding on surfaces when writing to final object
   bool rev_winding = !( mGreCustNoConform && o.mIsCustom );

   if( mGreCustNoConform && o.mIsCustom )
   {

      // We're not conforming the greeble to the surface, follow a different process to size/place the greeble

      // Size the greeble to fit inside the target surface's bounding box at its maximum size
      float ssizx = surf->xform_max.x - surf->xform_min.x;
      float ssizz = surf->xform_max.z - surf->xform_min.z;
      float gsizx = o.mMax.x - o.mMin.x;
      // Size based on the X coordinate
      float siz = ssizx / gsizx;
      for( unsigned int ix = 0; ix < o.mNumVerts; ++ix )
      {
         CGreVertex *v = &( o.mVerts[ix] );
         v->v.x *= siz;
         v->v.y *= siz;
         v->v.z *= siz;
      }
      o.FindBounds( true );
      // Size based on the Z coordinate, if larger than the X coordinate
      float gsizz = o.mMax.z - o.mMin.z;
      if( gsizz > ssizz )
      {
         siz = ssizz / gsizz;
         for( unsigned int ix = 0; ix < o.mNumVerts; ++ix )
         {
            CGreVertex *v = &( o.mVerts[ix] );
            v->v.x *= siz;
            v->v.y *= siz;
            v->v.z *= siz;
         }
         o.FindBounds( true );
      }

      // Now that the object "fits", reduce the area of our greeble's bounding box
      // to the target surface areas (if smaller)
      // Calculate the area of the surface
      float txarr[4], tyarr[4], tzarr[4];
      txarr[0] = o.mMin.x;
      txarr[1] = o.mMin.x;
      txarr[2] = o.mMax.x;
      txarr[3] = o.mMax.x;
      tyarr[0] = o.mMin.y;
      tyarr[1] = o.mMin.y;
      tyarr[2] = o.mMin.y;
      tyarr[3] = o.mMin.y;
      tzarr[0] = o.mMin.z;
      tzarr[1] = o.mMax.z;
      tzarr[2] = o.mMax.z;
      tzarr[3] = o.mMin.z;
      float g_surf_area = polygon_area_2_3d( 4, txarr, tyarr, tzarr );
      if( surf->surf_area < g_surf_area )
      {
         siz = surf->surf_area / g_surf_area;
         for( unsigned int ix = 0; ix < o.mNumVerts; ++ix )
         {
            CGreVertex *v = &( o.mVerts[ix] );
            v->v.x *= siz;
            v->v.y *= siz;
            v->v.z *= siz;
         }
         o.FindBounds( true );
      }

      // Keep reducing the size by 5% until all poly bounds fit.
      // If no poly bounds fit by the time we hit 5% of the original object size, just lock it in at 5%

      // Push surface verts into point2 array
      Point2 polyverts[5];
      for( int ix = surf->acsurf->numvert - 1; ix >= 0; --ix )
      {
         polyverts[ix].x = surf->xform_points[ix].x;
         polyverts[ix].y = surf->xform_points[ix].z;
      }
      polyverts[surf->acsurf->numvert].x = surf->xform_points[0].x;
      polyverts[surf->acsurf->numvert].y = surf->xform_points[0].z;
      float reduce_amt = 0.05f;
      float tot_siz_fac = 1.0f;
      Point2 p2min, p2max, p2[4];
      for( int it = 0; it < 19; ++it )
      {
         p2min.x = o.mMin.x * tot_siz_fac;
         p2min.y = o.mMin.z * tot_siz_fac;
         p2max.x = o.mMax.x * tot_siz_fac;
         p2max.y = o.mMax.z * tot_siz_fac;
         p2[0].x = p2min.x;
         p2[0].y = p2min.y;
         p2[1].x = p2min.x;
         p2[1].y = p2max.y;
         p2[2].x = p2max.x;
         p2[2].y = p2max.y;
         p2[3].x = p2max.x;
         p2[3].y = p2min.y;
         if( point_in_poly( polyverts, surf->acsurf->numvert, &( p2[0] )) &&
             point_in_poly( polyverts, surf->acsurf->numvert, &( p2[1] )) &&
             point_in_poly( polyverts, surf->acsurf->numvert, &( p2[2] )) &&
             point_in_poly( polyverts, surf->acsurf->numvert, &( p2[3] )))
         {
            break;
         }
         tot_siz_fac -= reduce_amt;
      }
      if( tot_siz_fac < 0.999999f )
      {
         for( unsigned int ix = 0; ix < o.mNumVerts; ++ix )
         {
            CGreVertex *v = &( o.mVerts[ix] );
            v->v.x *= tot_siz_fac;
            v->v.y *= tot_siz_fac;
            v->v.z *= tot_siz_fac;
         }
         o.FindBounds();
      }

      // Remember the original bounds
      Vertex oBoundsMin, oBoundsMax;
      memcpy( &oBoundsMin, &o.mMin, sizeof( oBoundsMin ));
      memcpy( &oBoundsMax, &o.mMax, sizeof( oBoundsMax ));

      // Resize the greeble
      siz = ( float )( DRAND * ( mGreSizePctMax - mGreSizePctMin ) + mGreSizePctMin );
      // Move points toward zero
      for( unsigned int ix = 0; ix < o.mNumVerts; ++ix )
      {
         CGreVertex *v = &( o.mVerts[ix] );
         v->v.x *= siz;
         v->v.y *= siz;
         v->v.z *= siz;
      }
      o.FindBounds();

      // Randomly locate the greeble, staying within the original bounds
      float maxmovx = ( oBoundsMax.x - o.mMax.x );
      float maxmovz = ( oBoundsMax.z - o.mMax.z );
      float movx = ( float )( DRAND * maxmovx * 2.0f ) - maxmovx;
      float movz = ( float )( DRAND * maxmovz * 2.0f ) - maxmovz;
      for( unsigned int ix = 0; ix < o.mNumVerts; ++ix )
      {
         CGreVertex *v = &( o.mVerts[ix] );
         v->v.x += movx;
         v->v.z += movz;
      }

   }
   else
   {

      // Follow the normal greeble process of conforming to the surface

      // Remember the original bounds
      Vertex oBoundsMin, oBoundsMax;
      memcpy( &oBoundsMin, &o.mMin, sizeof( oBoundsMin ));
      memcpy( &oBoundsMax, &o.mMax, sizeof( oBoundsMax ));

      // Resize the greeble
      float siz = ( float )( DRAND * ( mGreSizePctMax - mGreSizePctMin ) + mGreSizePctMin );
      // Move points toward zero
      for( unsigned int ix = 0; ix < o.mNumVerts; ++ix )
      {
         CGreVertex *v = &( o.mVerts[ix] );
         v->v.x *= siz;
         v->v.z *= siz;
      }
      o.FindBounds();

      // Randomly locate the greeble, staying within the original bounds
      float maxmovx = ( oBoundsMax.x - o.mMax.x );
      float maxmovz = ( oBoundsMax.z - o.mMax.z );
      float movx = ( float )( DRAND * maxmovx * 2.0f ) - maxmovx;
      float movz = ( float )( DRAND * maxmovz * 2.0f ) - maxmovz;
      for( unsigned int ix = 0; ix < o.mNumVerts; ++ix )
      {
         CGreVertex *v = &( o.mVerts[ix] );
         v->v.x += movx;
         v->v.z += movz;
      }

      // Match the shape of the target surface on the XZ plane
      for( unsigned int ix = 0; ix < o.mNumVerts; ++ix )
      {
         CGreVertex *v = &( o.mVerts[ix] );
         // Plot the coordinates of each vertex against the source surface.
         // Vertex 0 of our source surface is the "bottom right" of our XZ shape.
         // pctx, pctz  The x/z percentage of the current vertex's location within the bounds of the greeble.
         // vedge       Stores the XZ locations on the target surface edges used to plot the points.
         //             Serves as a sort of abcde... and 12345... coordinates on the edge of a map.
         //             The indices of this array are:
         //                 0 = bottom
         //                 1 = left
         //                 2 = top
         //                 3 = right
         float pctx, pctz;
         Point2 vedge[4];
         // Get the percentage of this vertex's location in the bounds of the greeble
         pctx = ( v->v.x - oBoundsMin.x ) / ( oBoundsMax.x - oBoundsMin.x );
         pctz = ( v->v.z - oBoundsMin.z ) / ( oBoundsMax.z - oBoundsMin.z );
         // Get the border locations
         vedge[0].x = bl->x + (( br->x - bl->x ) * pctx );
         vedge[0].y = bl->z + (( br->z - bl->z ) * pctx );
         vedge[1].x = br->x + (( tr->x - br->x ) * pctz );
         vedge[1].y = br->z + (( tr->z - br->z ) * pctz );
         vedge[2].x = tl->x + (( tr->x - tl->x ) * pctx );
         vedge[2].y = tl->z + (( tr->z - tl->z ) * pctx );
         vedge[3].x = bl->x + (( tl->x - bl->x ) * pctz );
         vedge[3].y = bl->z + (( tl->z - bl->z ) * pctz );
         // Get the intersection formed by the two line segments
         Point2 isect;
         isect_2d( vedge[0], vedge[2], vedge[1], vedge[3], &isect );
         // Plot the point to the intersection
         v->v.x = isect.x;
         v->v.z = isect.y;
      }

      // Scale the height of the greeble
      // Base height of the greeble is 200% of the surface area
      if( in_greeble->mYheight != 0.0f )
      {
         float gheight = surf->surf_area * ( float )( DRAND * ( mGreHeightPctMax - mGreHeightPctMin ) + mGreHeightPctMin );
         for( unsigned int ix = 0; ix < o.mNumVerts; ++ix )
         {
            CGreVertex *v = &( o.mVerts[ix] );
            v->v.y = ( v->v.y / in_greeble->mYheight ) * gheight;
         }
      }

   }

   // Rotate about XZ axis to match the last rotation we did to the CTargetSurface
   float rot_xz = -( surf->rot_xz );
   for( unsigned int ix = 0; ix < o.mNumVerts; ++ix )
   {
      float newx = o.mVerts[ix].v.x * cos( rot_xz ) - o.mVerts[ix].v.z * sin( rot_xz );
      float newz = o.mVerts[ix].v.x * sin( rot_xz ) + o.mVerts[ix].v.z * cos( rot_xz );
      o.mVerts[ix].v.x = newx;
      o.mVerts[ix].v.z = newz;
   }

   // Rotate to original surface's angle
   Vertex up;
   up.x = 0;
   up.y = 1;
   up.z = 0;
   Vertex norm;
   norm.x = surf->acsurf->normal.x;
   norm.y = surf->acsurf->normal.y;
   norm.z = surf->acsurf->normal.z;
   Vertex axis;
   C3d::VectorAxis( &up, &norm, &axis );
   float cosa = C3d::VectorCosAngle( &norm, &up );
   o.RotateAxis( &axis, cosa );

   // Move to the surface's original location
   for( unsigned int ix = 0; ix < o.mNumVerts; ++ix )
   {
      CGreVertex *v = &( o.mVerts[ix] );
      v->v.x += surf->surf_center.x;
      v->v.y += surf->surf_center.y;
      v->v.z += surf->surf_center.z;
   }

   ACObject *newobj = NULL;
   if( mObjKeepSeparate )
   {
      newobj = ac_new_object();
      ac_object_set_name( newobj, "greeble" );
      list_add_item_head( newobjs, newobj );
   }

   // Push vertices into final array
   for( unsigned int ix = 0; ix < o.mNumVerts; ++ix )
   {
      CGreVertex *v = &( o.mVerts[ix] );
      v->ac_v = new_vertex_set( v->v.x, v->v.y, v->v.z );
      list_add_item_head( newverts, v->ac_v );
      if( mObjKeepSeparate )
      {
         object_add_vertex_head( newobj, v->ac_v );
      }
   }
   // Push surfaces into final array
   for( unsigned int ix = 0; ix < o.mNumSurfs; ++ix )
   {
      CGreSurface *s = &( o.mSurfs[ix] );
      Surface *nsurf = new_surface();
      for( int iy = 0; iy < s->numverts; ++iy )
      {
         CGreSVertex *sv = &( s->sverts[iy] );
         if( rev_winding )
         {
            surface_add_vertex_head( nsurf, o.mVerts[sv->v].ac_v, sv->tx, sv->ty );
         }
         else
         {
            surface_add_vertex( nsurf, o.mVerts[sv->v].ac_v, sv->tx, sv->ty );
         }
      }
      surface_copy_properties( surf->acsurf, nsurf );
      list_add_item_head( newsurfs, nsurf );
      if( mObjKeepSeparate )
      {
         object_add_surface_head( newobj, nsurf );
      }
   }

   // Copy the texture index
   if( mObjKeepSeparate )
   {
      if( o.mTexIndex < 0 )
      {
         ACObject *parentobj = object_of_surface( surf->acsurf );
         if( ac_object_has_texture( parentobj ))
         {
            object_texture_set( newobj, ac_object_get_texture_index( parentobj ));
         }
      }
      else
      {
         object_texture_set( newobj, o.mTexIndex );
      }
   }

}
//----------------------------------------------------------------------

#undef IRAND
#undef DRAND_RANGE
